%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%% Instrumentation of the compiled Prolog code obtained when
%% we compile the original curry program into Prolog
%%
%%
%% We have coded the operator geq ('greater or equal than ..')
%% as follows:
%%
%% geq zero zero = True
%% geq (succ _) zero = True
%% geq zero (succ _) = False
%% geq (succ x) (succ y) = geq x y
%%
%% Although it is usually coded in this way (which is more efficient):
%%
%% geq2  _ zero = True
%% geq2 zero (succ _) = False
%% geq2 (succ x) (succ y) = geq2 x y
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Auxiliary predicate to built natural numbers.
built_nat(0, zero) :- !.
built_nat(I, succ(N)) :- NextI is (I - 1), built_nat(NextI, N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%
%%% block geq
geq(A,B,C):-hnf(A,F),geq_1(F,B,C).

geq_1(zero,A,B):-hnf(A,E),geq_1_zero_1(E,B).
geq_1_zero_1(zero,'True').
geq_1_zero_1(succ(_),'False').

geq_1(succ(A),B,C):-hnf(B,F),geq_1_succ_2(F,A,C).
geq_1_succ_2(zero,_,'True').
geq_1_succ_2(succ(A),B,C):-geq(B,A,C).



%%%%%%%%%%%%%%%%%%%%%%
%%% block IMPROVED geq
igeq(A,B,C):-hnf(A,F), hnf(B,G), igeq_1(F,G,C).

igeq_1(zero,zero,'True').
igeq_1(zero,succ(_),'False').
igeq_1(succ(_),zero,'True').
igeq_1(succ(A),succ(B),C):- igeq(A,B,C).


%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(geq(A,B),C):-!,geq(A,B,C).
hnf(igeq(A,B),C):-!,igeq(A,B,C).
hnf(A,A).

%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf('True','True'):-!.
nf_hnf('False','False'):-!.
nf_hnf(zero,zero):-!.
nf_hnf(succ(A),succ(B)):-!,nf(A,B).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%/************** SICSTUS
% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%
g0(T) :- write('goal 0: ...'), stat_start,
      nf(_, _),
      stat_end(T),
      nl.



%%%%%
g1(T) :- write('goal 1 (100000 >= 99999):  geq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start,
      nf(geq(N2, N1), _),
      stat_end(T),
      nl.

%%%%%
g2(T) :- write('goal 2 (100000 >= 99999):  igeq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start,
      nf(igeq(N2, N1), _),
      stat_end(T),
      nl.



%%%%%
g11(T) :- write('goal 1 (100000 < 99999):  geq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start,
      nf(geq(N1, N2), _),
      stat_end(T),
      nl.

%%%%%
g21(T) :- write('goal 2 (100000 < 99999):  igeq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start,
      nf(igeq(N1, N2), _),
      stat_end(T),
      nl.

%%%*********************************************************************/

/**************** SWI-ROLOG
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% T1 is the start time
stat_start(T1) :- T1 is cputime.

%% stat_end(Input,Output)
%% T1 is the starting time and T is the temporal cost of the whole computation,
stat_end(T1,T) :-
	T2 is cputime, T is T2-T1, write('T= '), write(T),
        nl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%
g0(T) :- write('goal 0: ...'), stat_start(T1),
      nf(_, _),
      stat_end(T1,T).


%%%%%
g1(T) :- write('goal 1 (100000 >= 99999):  geq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start(T1),
      nf(geq(N2, N1), _),
      stat_end(T1,T).

%%%%%
g2(T) :- write('goal 2 (100000 >= 99999):  igeq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start(T1),
      nf(igeq(N2, N1), _),
      stat_end(T1,T).



%%%%%
g11(T) :- write('goal 1 (100000 < 99999):  geq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start(T1),
      nf(geq(N1, N2), _),
      stat_end(T1,T).

%%%%%
g21(T) :- write('goal 2 (100000 < 99999):  igeq ... '),
      built_nat(99999, N1),   built_nat(100000, N2),
      stat_start(T1),
      nf(igeq(N1, N2), _),
      stat_end(T1,T).



*********************************************************************/


allgoals:- g0(T0),g1(T1),g2(T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl, nl,
		g0(T01),g11(T11),g21(T21),
		Sup1 is (((T11-T01)-(T21-T01))/(T11-T01))*100,
		write('Speedup = '), write(Sup1), write('%'), nl.


