%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
%%  R_1: f(zero, zero) ---> zero, 
%%  R_2: f(succ(X), zero) --->  succ(zero),
%%  R_3: f(succ(X), succ(succ(Y))) --->  f(X, Y).
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Auxiliary predicate to built natural numbers.
built_nat(0, zero) :- !.
built_nat(I, succ(N)) :- NextI is (I - 1), built_nat(NextI, N).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block f : CURRY2PROLOG
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Clause for the root node: it exploits the first inductive position
f(X1, X2, H) :- hnf(X1, HX1), f_1(HX1, X2, H).

% Clauses for the remainder nodes:
f_1(zero, X2, H) :- hnf(X2, HX2), f_1_zero_2(HX2, H).
f_1_zero_2(zero, zero).

f_1(succ(X3), X2, H) :- hnf(X2, HX2), f_1_succ_2(HX2, X3, H).
f_1_succ_2(zero, _, succ(zero)).
f_1_succ_2(succ(X4), X3, H):- hnf(X4, HX4), f_1_succ_2_succ_21(HX4, X3, H).
f_1_succ_2_succ_21(succ(X5), X3, H):- f(X3, X5, H).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block IMPROVED f 
%% (exploiting the deterministic branch)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Clause for the root node: it exploits the first inductive position
impf(X1, X2, H) :- hnf(X1, HX1), impf_1(HX1, X2, H).

% Clauses for the remainder nodes:
impf_1(zero, X2, H) :- hnf(X2, HX2), impf_1_zero_2(HX2, H).
impf_1_zero_2(zero, zero).

impf_1(succ(X3), X2, H) :- hnf(X2, HX2), impf_1_succ_2(HX2, X3, H).
impf_1_succ_2(zero, _, succ(zero)).
impf_1_succ_2(succ(X4), X3, H):- hnf(X4, succ(X5)), impf(X3, X5, H).





%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(f(A,B),D):-!,f(A,B,D).
hnf(impf(A,B),D):-!,impf(A,B,D).
hnf(A,A).

%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf(zero,zero):-!.
nf_hnf(succ(A),succ(B)):-!,nf(A,B).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%% SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
test1(H) :- built_nat(4, N1), built_nat(8, N2), nf(f(N1,N2), H).
test2(H) :- built_nat(5, N1), built_nat(8, N2), nf(f(N1,N2), H).
test11(H) :- built_nat(4, N1), built_nat(8, N2), nf(impf(N1,N2), H).
test22(H) :- built_nat(5, N1), built_nat(8, N2), nf(impf(N1,N2), H).

%%%%%
g0(T) :- write('goal 0: ...'), stat_start,
      nf(_, _),
      stat_end(T),
      nl.


%%%%%
g1(T) :- write('goal 1:  f(50000, 100000) ... '),
      built_nat(50000, N1), built_nat(100000, N2),
      stat_start,
      nf(f(N1,N2), _),
      stat_end(T),
      nl.

%%%%%
g2(T) :- write('goal 1:  impf(50000, 100000) ... '),
      built_nat(50000, N1), built_nat(100000, N2),
      stat_start,
      nf(impf(N1,N2), _),
      stat_end(T),
      nl.

%%%%%
g3(T) :- write('goal 3:  f(100000, 50000) ... '),
      built_nat(100000, N1), built_nat(50000, N2),
      stat_start,
      nf(f(N1,N2), _),
      stat_end(T),
      nl.

%%%%%
g4(T) :- write('goal 4:  impf(100000, 50000) ... '),
      built_nat(100000, N1), built_nat(50000, N2),
      stat_start,
      nf(impf(N1,N2), _),
      stat_end(T),
      nl.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
allgoals:- g0(T0),g1(T1),g2(T2),
		Sup is (((T1-T0)-(T2-T0))/(T1-T0))*100,
		write('Speedup = '), write(Sup), write('%'), nl, nl,
		g3(T3),g4(T4),
		Sup1 is (((T3-T0)-(T4-T0))/(T3-T0))*100,
		write('Speedup = '), write(Sup1), write('%'), nl, nl.





