%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The experiments contained in this file was designed by 
%% Gines Moreno-Valverde in order to show how different unfolding
%% strategies may affect the efficiency of the Prolog code generated
%% in the compilation of the (partial) function even:
%%
%%  R_1: even(0) ---> 'True', 
%%  R_2: even(succ(succ(X)) --->  even(X).
%%
%% We gratefully acknowledge Gines Moreno-Valverde for his valuable
%% contribution.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Auxiliary predicate to built natural numbers.
built_nat(0, zero) :- !.
built_nat(I, succ(N)) :- NextI is (I - 1), built_nat(NextI, N).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  block HANUS95 even 
%%
%% All results are compared with regard to the standard translation
%% technique described in "Efficient translation of lazy functional 
%% logic programs into Prolog" (In Proc. LOPSTR'95, pages 252--266. 
%% Springer LNCS 1048, 1995) by M. Hanus, 
%% that we call "HANUS95". Note that, in this original translation 
%% technique, the final unfolding of rule nodes is not performed, 
%% losing an opportunity for determinate unfolding. Actual implementation 
%% of the curry2prolog compiler takes advantage of this possibility.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Clause for the root node: it exploits the first inductive position
even(X1, H) :- hnf(X1, HX1), even_1(HX1, H).

% Clauses for the remainder nodes:
even_1(zero, 'True').
even_1(succ(X2), H) :- hnf(X2, HX2), even_1_s_1_1(HX2, H).

even_1_s_1_1(succ(X3), H) :- hnf(even(X3), H).


%%%%%%%%%%%%%
%%% block hnf
hnf(A,B):-var(A),!,B=A.
hnf(even(A),D):-!,even(A,D).
hnf(A,A).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even:  UNFOLDING 1
%%
%% We get a clause of the Prolog code HANUS95 as the unfolded clause 
%% and we select an atom of the form "even_*"  in its body to
%% perform a DETERMINATE unfolding step. 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Clause for the root node: it exploits the first inductive position
u1_even(X1, H) :- u1_hnf(X1, HX1), u1_even_1(HX1, H).

% Clauses for the remainder nodes:
u1_even_1(zero, 'True').
u1_even_1(succ(X2), H) :- u1_hnf(X2, succ(X3)),  u1_hnf(u1_even(X3), H).


%%%%%%%%%%%%%
%%% block u1_hnf
u1_hnf(A,B):-var(A),!,B=A.
u1_hnf(u1_even(A),D):-!,u1_even(A,D).
u1_hnf(A,A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even:  UNFOLDING 2
%%
%% We get a clause of the Prolog code HANUS95 as the unfolded clause 
%% and we select an atom of the form "hnf"  in its body to
%% perform a DETERMINATE unfolding step. In this example, the only 
%% possibility is "hnf(even(X3), H)". This corresponds with the
%% DETERMINATE unfolding of rule nodes performed in actual implementations 
%% of the curry2prolog compiler.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Clause for the root node: it exploits the first inductive position
u2_even(X1, H) :- u2_hnf(X1, HX1), u2_even_1(HX1, H).

% Clauses for the remainder nodes:
u2_even_1(zero, 'True').
u2_even_1(succ(X2), H) :- u2_hnf(X2, HX2), u2_even_1_s_1_1(HX2, H).

u2_even_1_s_1_1(succ(X3), H) :- u2_even(X3, H).

%%%%%%%%%%%%%
%%% block u2_hnf
u2_hnf(A,B):-var(A),!,B=A.
u2_hnf(u2_even(A),D):-!,u2_even(A,D).
u2_hnf(A,A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even : UNFOLDING 3
%%
%% We get a clause of the Prolog code HANUS95 as the unfolded clause 
%% and we select an atom of the form "even_*"  in its body to
%% perform an unfolding step (possibly NON DETERMINATE).
%% This experiment shows that in our context, NON DETERMINATE 
%% unfolding is harmful.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
% Two clauses for the root node. 
u3_even(X1,'True'):-u3_hnf(X1,zero).
u3_even(X1,H):-u3_hnf(X1,succ(X2)),u3_hnf(X2,HX2),u3_even_1_s_1_1(HX2, H).


% Clauses for the remainder nodes:
u3_even_1_s_1_1(succ(X3), H) :- u3_hnf(u3_even(X3), H).

%%%%%%%%%%%%%
%%% block u3_hnf
u3_hnf(A,B):-var(A),!,B=A.
u3_hnf(u3_even(A),D):-!,u3_even(A,D).
u3_hnf(A,A).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even: UNFOLDING 1 + UNFOLDING 2 
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
% Clause for the root node: it exploits the first inductive position
u1_u2_even(X1, H) :- u1_u2_hnf(X1, HX1), u1_u2_even_1(HX1, H).

% Clauses for the remainder nodes:
u1_u2_even_1(zero, 'True').
u1_u2_even_1(succ(X2), H) :- u1_u2_hnf(X2, succ(X3)),!,u1_u2_even(X3, H).


%%%%%%%%%%%%%
%%% block u1_u2_hnf
u1_u2_hnf(A,B):-var(A),!,B=A.
u1_u2_hnf(u1_u2_even(A),D):-!,u1_u2_even(A,D).
u1_u2_hnf(A,A).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even: UNFOLDING 1 + UNFOLDING 3
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%

u1_u3_even(X1,'True'):-u1_u3_hnf(X1,zero).
u1_u3_even(X1,H):-u1_u3_hnf(X1,succ(X2)),
                  u1_u3_hnf(X2,succ(X3)),u1_u3_hnf(u1_u3_even(X3), H).



%%%%%%%%%%%%%
%%% block u1_u3_hnf
u1_u3_hnf(A,B):-var(A),!,B=A.
u1_u3_hnf(u1_u3_even(A),D):-!,u1_u3_even(A,D).
u1_u3_hnf(A,A).




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% block even: UNFOLDING 1 + UNFOLDING 2 + UNFOLDING 3
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%
% Two clauses for the whole function symbol

u1_u2_u3_even(X1,'True'):-u1_u2_u3_hnf(X1,zero).
u1_u2_u3_even(X1,H):-u1_u2_u3_hnf(X1,succ(X2)),
                     u1_u2_u3_hnf(X2,succ(X3)),!,u1_u2_u3_even(X3, H).



%%%%%%%%%%%%%
%%% block u1_u2_u3_hnf
u1_u2_u3_hnf(A,B):-var(A),!,B=A.
u1_u2_u3_hnf(u1_u2_u3_even(A),D):-!,u1_u2_u3_even(A,D).
u1_u2_u3_hnf(A,A).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%
%%% block nf
nf(A,B):-hnf(A,E),nf_hnf(E,B).

%%% block nf_hnf
nf_hnf(A,B):-var(A),!,B=A.
nf_hnf('True','True'):-!.
nf_hnf('False','False'):-!.
nf_hnf(zero,zero):-!.
nf_hnf(succ(A),succ(B)):-!,nf(A,B).



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%% Example of goals considered in the experiments %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%% SICSTUS
%%%% This is how time and space are measured.
stat_start :- garbage_collect, nogc, statistics(runtime,_).
stat_end(Time) :-
        statistics(runtime,[_,MSec]), Time is MSec/1000,
        statistics(global_stack,[Bytes,_]), Mb is Bytes/1000000,
        width(WT,Time), format("Time ~*c~2f sec   ", [WT, 0' , Time]),
        width(WM,Mb), format("Garb ~*c~3f Mb   ", [WM, 0' , Mb]),
        nl.

% statistics
width(0,X) :- X>=10, !.
width(1,_).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

goal(G,T0,T1):-G=..[H|_],write(H), write(':                      '),
      stat_start,
      call(G),
      stat_end(T),
      Sup is (((T1-T0)-(T-T0))/(T1-T0))*100,
	write('Speedup = '), write(Sup), write('%'), nl,nl.


allgoals(X):-  built_nat(X, N1), 
            write('Initializing................'),
            stat_start,nf(_,_),stat_end(T0), 
		write('HANUS95 even:              '),
            stat_start,even(N1,_),stat_end(T1), nl,
     		goal(u1_even(N1,_),T0,T1),
		goal(u2_even(N1,_),T0,T1),
		goal(u3_even(N1,_),T0,T1),
		goal(u1_u2_even(N1,_),T0,T1),      
           	goal(u1_u3_even(N1,_),T0,T1),
           	goal(u1_u2_u3_even(N1,_),T0,T1). 
            
           

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%             SUMMARY                               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
/*-----------------------------------------------------
| ?- allgoals(100000).
Initializing................Time  0.00 sec   Garb  0.802 Mb
HANUS95 even:              Time  0.65 sec   Garb 24.002 Mb

u1_even:                      Time  0.61 sec   Garb 21.202 Mb
Speedup = 6.153846153846159%

u2_even:                      Time  0.48 sec   Garb 18.402 Mb
Speedup = 26.15384615384616%

u3_even:                      Time  0.76 sec   Garb 21.802 Mb
Speedup = -16.92307692307692%

u1_u2_even:                      Time  0.46 sec   Garb 17.002 Mb
Speedup = 29.230769230769234%

u1_u3_even:                      Time  0.70 sec   Garb 19.002 Mb
Speedup = -7.692307692307682%

u1_u2_u3_even:                      Time  0.55 sec   Garb 14.802 Mb
Speedup = 15.38461538461538%

yes
------------------------------------------------------------*/
